'''
SYNOPSIS

    class RecvHandler

DESCRIPTION

    Receive request HTTP message, initial req and res
'''

from .http_error import HttpArgumentError, Http400Error
from .mime_type import MIME_JSON
from .ihandler import IRecvHandler


class RecvHandler(IRecvHandler):
    def handle(self, req, res, env):
        res.status = 200
        res.content_type = MIME_JSON
        res.header = {}

        req.method = env['REQUEST_METHOD']
        req.path = env['PATH_INFO']
        req.header = self._parse_header(env)
        req.server_name = env['SERVER_NAME']
        req.server_port = int(env['SERVER_PORT'])
        req.server_protocol = env['SERVER_PROTOCOL']
        req.remote_addr = env['REMOTE_ADDR']
        if req.method.lower() not in ['get', 'head', 'option']:
            req.content_type = env['CONTENT_TYPE']
        try:
            req.args = self._parse_args(env['QUERY_STRING'])
        except HttpArgumentError:
            raise Http400Error(req, 'query string is invalid')

        # read all of body message to memory
        req.content_length = 0
        try:
            req.content_length = int(env.get('CONTENT_LENGTH', 0))
        except ValueError:
            pass
        if req.content_length > 0:
            req.body = env['wsgi.input'].read(req.content_length)

    def _parse_args(self, arg_str):
        args = {}
        if arg_str is None or len(arg_str) == 0:
            return args
        arg_coms = arg_str.split('&')
        for arg_com in arg_coms:
            arg_parts = arg_com.split('=')
            if len(arg_parts) != 2:
                raise HttpArgumentError(arg_str)
            if len(arg_parts[0]) == 0 or len(arg_parts[1]) == 0:
                raise HttpArgumentError(arg_str)
            args[arg_parts[0]] = arg_parts[1]
        return args

    def _parse_header(self, env):
        header = {}
        for key, value in env.items():
            if key[:5] != 'HTTP_':
                continue
            header[key[5:]] = value
        return header



import json

from .http_error import Http400Error
from .ihandler import IHandler
from .encoding import UTF_8
from .mime_type import MIME_JSON


class ReqJsonHandler(IHandler):
    def handle(self, req, res):
        if req.content_type != MIME_JSON:
            return
        if req.body is None:
            return
        try:
            req.body = json.loads(req.body.decode(UTF_8))
        except ValueError:
            raise Http400Error(req, 'body is invalid json format')



import logging
from os.path import isfile

from .ihandler import IErrorHandler
from .logging import LOGFILE_MODE
from .shell import touch


class ErrorLogHandler(IErrorHandler):
    def __init__(self, file):
        self._file = file

        if not isfile(file):
            touch(file, LOGFILE_MODE)

        self._logger = logging.getLogger(file)
        fhandler = logging.FileHandler(file)
        formatter = logging.Formatter('%(asctime)s$ %(message)s')
        fhandler.setFormatter(formatter)
        self._logger.addHandler(fhandler)
        self._logger.setLevel(logging.INFO)

    @property
    def file(self):
        return self._file

    def handle(self, req, res, e):
        msg = ' '.join([
            str(res.status), req.remote_addr, req.method, req.path,
            type(e).__name__
        ])
        self._logger.info(msg)
        return True



'''
SYNOPSIS

    interface IHandler
    interface IErrorHandler

DESCRIPTION

    Describe interface allow process pipeline.
'''

from abc import ABC, abstractmethod


class IHandler(ABC):
    @abstractmethod
    def handle(self, req, res):
        pass


class IErrorHandler(ABC):
    @abstractmethod
    def handle(self, req, res, e):
        pass


class IRecvHandler(ABC):
    @abstractmethod
    def handle(self, req, res, wsgi_env):
        pass


class ISendHandler(ABC):
    @abstractmethod
    def handle(self, req, res, wsgi_send):
        pass



#!/usr/bin/env python

'''
SYNOPSIS

    pip install -e .

DESCRIPTION

    Metadata for this package.

REFERENCES

    Building and Distributing Packages with Setuptools
        http://setuptools.readthedocs.io/en/latest/setuptools.html
'''

from setuptools import setup, find_packages


setup(
    name='clink',
    version='0.7.0',
    description='HTTP APIs Framework',
    keywords='http api framework',
    author='Kevin Leptons',
    author_email='kevin.leptons@gmail.com',
    url='https://github.com/kevin-leptons/clink',
    download_url='https://github.com/kevin-leptons/clink',
    python_requires=">=3.4",
    install_requires=[
    ],
    packages=find_packages(exclude=['tool', 'test']),
)



from os import path

from .type import Request, Response
from .recv_handler import RecvHandler
from .send_handler import SendHandler
from .req_json_handler import ReqJsonHandler
from .req_urlenc_handler import ReqUrlEncHandler
from .req_log_handler import ReqLogHandler
from .res_json_handler import ResJsonHandler
from .res_cors_handler import ResCorsHandler
from .err_http_handler import ErrorHttpHandler
from .err_log_handler import ErrorLogHandler


class Application():
    req_log_handler = None
    err_log_handler = None

    req_handlers = None
    res_handlers = None
    err_handlers = None

    def __init__(self, name, router):
        self._name = name
        self._router = router

        self.recv_handler = RecvHandler()
        self.send_handler = SendHandler()

        req_logfile = path.join('/var/tmp', name, 'request.log')
        self.req_log_handler = ReqLogHandler(req_logfile)

        err_logfile = path.join('/var/tmp', name, 'error.log')
        self.err_log_handler = ErrorLogHandler(err_logfile)

        self.err_handlers = [ErrorHttpHandler()]
        self.req_handlers = [ReqJsonHandler(), ReqUrlEncHandler()]
        self.res_handlers = [ResJsonHandler(), ResCorsHandler()]

    def __call__(self, wsgi_env, wsgi_send):
        req = Request()
        res = Response()

        try:
            # receive message, initial req and res
            self.recv_handler.handle(req, res, wsgi_env)

            # perform request logging handler
            self.req_log_handler.handle(req, res)

            # perform request handlers
            for handler in self.req_handlers:
                handler.handle(req, res)

            # routing, find main handler and perform it
            m_req_handler = self._router.find_route_handler(req)
            m_req_handler(req, res)

            # perform response handlers
            for handler in self.res_handlers:
                handler.handle(req, res)

            # send response
            return self.send_handler.handle(req, res, wsgi_send)
        except Exception as e:
            # perform error handlers
            handled = False
            for handler in self.err_handlers:
                if handler.handle(req, res, e):
                    handled = True
            if not handled:
                raise e

            # send error response
            return self.send_handler.handle(req, res, wsgi_send)

            # perform error logging handler
            self.err_log_handler.handle(req, res, e)



"""
SYNOPSIS

    class RouteExistError
    class RoutePathError

    class Route
    class RouteNode
    class Router

DESCRIPTION

    Detect request handle and url parameters. It's not support parameters in
    url for performace, let use arugments.
"""

import os

from .http_error import Http404Error, Http405Error, Http406Error
from .mime_type import MIME_JSON

_URL_SLASH = '/'
_PARAM_CHAR = ':'


class RouteExistError(Exception):
    def __init__(self, spec):
        self.spec = spec
        self._msg = '%s %s; %s' % (
            spec.method, spec.path, spec.req_type
        )

    def __str___(self):
        return self._msg


class RoutePathError(Exception):
    def __init__(self, path):
        self.path = path

    def __str___(self):
        return self.path


class RouteNode():
    def __init__(self, name):
        self.name = name
        self.child = {}
        self.specs = []


class RouteSpec():
    def __init__(self, method, path, handler, req_type):
        self.method = method
        self.path = path
        self.handler = handler
        self.req_type = req_type


class Route():
    def __init__(self, base_path):
        if not self.verify_path(base_path):
            raise RoutePathError(base_path)
        self.base_path = base_path
        self.specs = []

    def get(self, path):
        return self.map('get', path, None)

    def post(self, path, req_type=MIME_JSON):
        return self.map('post', path, req_type)

    def put(self, path, req_type=MIME_JSON):
        return self.map('put', path, req_type)

    def patch(self, path, req_type=MIME_JSON):
        return self.map('patch', path, req_type)

    def delete(self, path):
        return self.map('delete', path, None)

    def option(self, path):
        return self.map('option', path, None)

    def map(self, method, path, req_type):
        if not self.verify_path(path):
            raise RoutePathError(path)

        def decorator_fn(target_fn):
            abs_path = os.path.join(self.base_path, path)
            spec = RouteSpec(method, abs_path, target_fn, req_type)
            self.specs.append(spec)
        return decorator_fn

    def verify_path(self, path):
        if len(path) > 0:
            if path[0] == _URL_SLASH:
                return False
        return True


class Router:
    def __init__(self, routes):
        self._root_route = RouteNode('/')
        for route in routes:
            for spec in route.specs:
                self.add_route(spec)

    def find_route_handler(self, req):
        node = self.find_node(req.path)
        if node is None or len(node.specs) == 0:
            raise Http404Error(req)
        m_ok_specs = [s for s in node.specs if s.method == req.method.lower()]
        if len(m_ok_specs) == 0:
            raise Http405Error(req)
        for spec in m_ok_specs:
            if spec.req_type == req.content_type:
                return spec.handler
        raise Http406Error(req)

    def add_route(self, spec):
        # search and add nodes
        node = self._root_route
        if spec.path != _URL_SLASH:
            node_names = spec.path.split(_URL_SLASH)
            node_names = list(filter(''.__ne__, node_names))
            for node_name in node_names:
                if node_name not in node.child:
                    node.child[node_name] = RouteNode(node_name)
                node = node.child[node_name]

        # add handle
        if self._spec_is_exist(node, spec):
            raise RouteExistError(spec)
        node.specs.append(spec)

    def find_node(self, path):
        node = self._root_route
        if path != _URL_SLASH:
            node_names = path.split(_URL_SLASH)
            node_names = list(filter(''.__ne__, node_names))
            for node_name in node_names:
                if node_name not in node.child:
                    return None
                node = node.child[node_name]
        return node

    def _spec_is_match(self, spec_1, spec_2):
        if spec_1.method != spec_2.method:
            return False
        if spec_1.req_type != spec_2.req_type:
            return False
        return True

    def _spec_is_exist(self, node, spec):
        for node_spec in node.specs:
            if self._spec_is_match(node_spec, spec):
                return True
        return False
